#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <string.h>
#include <utime.h>
#include <dirent.h>
#include <stdio.h>
#include <fcntl.h>
#include "../src/common.h"

/* ------------------------------------------------------------- **
** This is the cache manager bit. Each file in the cache has a
** cache_entry structure which is held in a hash table. The end of the
** filename should be well distributed and is used as the hashing key. 
** Each cache_entry structure also has a pair of pointers, and so can
** also be accessed as a doubly linked list with most recently
** accessed files nearer the head.
** 
** A running tally is kept of total file size, and the LRU cache file
** removed whenever this exceeds the maximum. 
**
** The only interface atm. is sending a single char + filename over
** the socket. "N" adds a new entry or updates an existing one, "X"
** deletes an entry, and "A" brings an entry to the top of the LRU
** list. 
** ------------------------------------------------------------- */

struct cache_entry {
	struct cache_entry *next, *prev, *collision;
	
	time_t last_access;
	time_t created;
	char filename[256];
	char uri[256];
	int size;
};

int cachemgr_init(char *cd, int cs);
void scan_dir(const char *dir);
struct cache_entry *make_entry(char *name);
struct cache_entry **hash_loc(char *key);
void add_entry(struct cache_entry *entry);
void purge_entry(struct cache_entry *entry);
void remove_entry(struct cache_entry *entry);

#define table_size 256

static struct cache_entry *hash_table[table_size];
static struct cache_entry *head=NULL, *tail=NULL;
static unsigned long total_size=0;
static int cachesize;
static char *cachedir;

int cachemgr_init(char *cd, int cs)
{
	struct cache_entry *p;
	FILE *fp;
	int fds[2];
	
	cachedir=cd;
	cachesize=cs;

	if(pipe(fds)==-1) return(-1);
	switch(fork()){
	case -1:
		close(fds[0]);
		close(fds[1]);
		return(-1);
	case 0:
		break;
	default:
		close(fds[0]);
		return(fds[1]);
	}
	close(fds[1]);
	scan_dir(cachedir);

	fp=fdopen(fds[0], "r");
	if(fp==NULL) {
		debug("Arrghh\n");
		exit(1);
	}
	do{
		char buf[MAX_LINE_LEN];
		
		if(fgets(buf, MAX_LINE_LEN, fp)==NULL) exit(0);
		buf[strlen(buf)-1]=0; /*FIXME*/
		switch(*buf){
		case 'N':
			debug2("Adding entry %s\n", buf+1);
			add_entry(make_entry(buf+1));
			break;
		case 'X':
			debug2("Deleting entry %s\n", buf+1);
			purge_entry(*hash_loc(buf+1));
			break;
		case 'A':
			debug2("LRUing entry %s\n", buf+1);
			p=*hash_loc(buf+1);
			
			if(p->next) p->next->prev=p->prev;
			else tail=p->prev;
			if(p->prev) p->prev->next=p->next;
			else head=p->next;
			
			p->next=head;
			head->prev=p;
			p->prev=NULL;
			head=p;
			break;
		default:
		}

	}while(1);
}

void scan_dir(const char *dir)
{
	DIR *dp;
	struct dirent *entry;
	struct stat status;
	char name[512];

	dp=opendir(dir);
	if(dp==NULL) {
		perror("opendir");
		exit(-1);
	}
	chdir(dir);
	while((entry = readdir(dp)) !=NULL) {
		if(entry->d_name[0]=='.') continue;
		stat(entry->d_name, &status);
		if(S_ISDIR(status.st_mode)) scan_dir(entry->d_name);
		else if (S_ISREG(status.st_mode)) {
			struct cache_entry *tmp;

			getcwd(name, 512);
			strcat(name, "/");
			strcat(name, entry->d_name);
			tmp=make_entry(name);
			if(tmp) add_entry(tmp);
		}
	}
	chdir("..");
	closedir(dp);
}

struct cache_entry *make_entry(char *name)
{
	struct stat status;
	struct cache_entry *ret;
	struct utimbuf tbuf;
	char buf[1024];
	int fd, i;

	ret = (struct cache_entry *) malloc(sizeof(struct cache_entry));
	if(ret==NULL) {
		perror("malloc");
		exit(-1);
	}

	stat(name, &status);
	tbuf.actime=status.st_atime;
	tbuf.modtime=status.st_mtime;

	strcpy(ret->filename,name);
	ret->last_access=status.st_atime;

	fd=open(ret->filename, O_RDONLY);
	read(fd, buf, 5);
	i=atoi(buf);
	if(read(fd, buf, i)==i){
		strtok(buf, " ");/*MDTM*/
		ret->size=atoi(strtok(NULL, " "));
		strtok(NULL, " ");/*Type*/
		strcpy(ret->uri, strtok(NULL, " \r\n"));
		close(fd);
	} else {
		close(fd);
		debug2("Invalid file %s\n", name);
		unlink(name);
		free(ret);
		return(NULL);
	}
	if(ret->size!=status.st_size-i-5){
		debug2("Partial file %s\n", name);
	}

	utime(name, &tbuf);
	return(ret);
}

struct cache_entry **hash_loc(char *key)
{
	int i;
	struct cache_entry *p;
	char *tmp;

	tmp=key+strlen(key)-4;
	i=strtol(tmp, NULL, 16) % table_size;
	
	if(hash_table[i]==NULL) return(&hash_table[i]);
	if(!strcmp(hash_table[i]->filename, key)) return(&hash_table[i]);

	debug("X");/*Collision*/
	for(p=hash_table[i]; 
	    p->collision!=NULL && strcmp(p->collision->filename, key); 
	    p=p->collision);
	return(&p->collision);
}

void add_entry(struct cache_entry *entry)
{
	struct cache_entry *p, **pp;

	entry->collision=NULL;

	pp=hash_loc(entry->filename);

	if(*pp!=NULL) {
		debug("   Already got it - removing old entry.\n");

		remove_entry(*pp);
		pp=hash_loc(entry->filename);
	}
	*pp=entry;

	/*Now add to linked list, based on last access time*/
	for(p=head;p!=NULL && entry->last_access < p->last_access; p=p->next);
	entry->next=p;
	entry->prev = p ? p->prev : tail;
	if(entry->prev) entry->prev->next=entry;
	else head=entry;
	if(entry->next) entry->next->prev=entry;
	else tail=entry;

	total_size+=entry->size;
	if(cachesize) while(total_size > cachesize) purge_entry(tail);	
}

void purge_entry(struct cache_entry *entry)
{
	unlink(entry->filename);
	remove_entry(entry);
}

void remove_entry(struct cache_entry *entry)
{
	int index;
	struct cache_entry *p;
	char *tmp;

	if(entry->next) entry->next->prev=entry->prev;
	else tail=entry->prev;
	if(entry->prev) entry->prev->next=entry->next;
	else head=entry->next;

	tmp=entry->filename+strlen(entry->filename)-4;
	index=strtol(tmp, NULL, 16) % table_size;

	debug2("Removing file %s\n", entry->uri);
	if(hash_table[index]==NULL){
		debug("Arrghhhhhhhhhhhhhhhhh\n");
		return;
	}

	if(entry==hash_table[index]){
		hash_table[index]=entry->collision;
	} else {
		for(p=hash_table[index]; 
		    p->collision!=NULL && p->collision!=entry; 
		    p=p->collision);
		if(p->collision==NULL) {
			debug("Removing non existant entry!\n");
			exit(-1);
		}
		p->collision=entry->collision;
	}

	total_size-=entry->size;
	free(entry);
}

